/*

	Copyright  2003 Frank Condello
	
	Portions Copyright  2001 Dair Grant <mailto:dair@webthing.net>
	and Copyright  2003 Joe Strout <joe@strout.net> (from "LoadSaveJPEG")
	
	"GetData" based on code by Will Leshner <leshner@ljug.com>
	
	Linux port contributed by Jonathan Johnson <jonj@realsoftware.com>

*/

/* PNGuMac.cpp: Mac OS routines */

#include "PNGu.h"
#include <strings.h>

#if defined(__MWERKS__)
# include <FSp_fopen.h>
#  if TARGET_API_MAC_CARBON
#   include "MachoGlue.h"
#  endif
#elif defined(__GNUC__)
# include <sys/param.h>
#endif

#define PNGUMAXPATH 2048

#if !TARGET_RT_MAC_MACHO
static Boolean _BuildImageDataFromPictureAndMaskQD(REALpicture thePicture, REALpicture theMask, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData);
static Boolean _BuildImageDataFromPictureQD(REALpicture thePicture, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData);
#endif
#if TARGET_API_MAC_CARBON
static Boolean _BuildImageDataFromPictureAndMaskCG(REALpicture thePicture, REALpicture theMask, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData);
static Boolean _BuildImageDataFromPictureCG(REALpicture thePicture, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData);
# if defined(__MWERKS__)
static SInt32 _isOSX(void);
# endif
#endif


#pragma mark -


/* ---------------------------------------------------------------------------
	OpenFILEFromFolderItem
	
	Open a FILE* for a FolderItem
	NOTE: CFM targets require Metrowerk's FSRef_fopen
---------------------------------------------------------------------------- */
FILE *OpenFILEFromFolderItem(REALfolderItem theFile, const char *fileMode)
{
	FSSpec			theFSSpec;
	FSRef			theFSRef;
	FSCatalogInfo	catInfo;
	FileInfo		fInfo;
	HFSUniStr255	uname;
	FILE			*ansiFile = NULL;
	char			*thePath = NULL;
	
	// Grab the FSSpec
	if (!REALFSSpecFromFolderItem(&theFSSpec, theFile))
		return(NULL);

	// And open the file...

	// If we're writing and the file doesn't exist, we need to create it first
	if ( strcmp(fileMode, "wb") == 0 && FSpMakeFSRef(&theFSSpec, &theFSRef) == fnfErr )
	{
		if ( REALFSRefFromFolderItem(theFile, &theFSRef, &uname) )
		{
			// File type is always PNGf, but creator is null on OS X (let Launch Services handle it)
			BlockZero(&fInfo, sizeof(fInfo));
			BlockZero(&catInfo, sizeof(catInfo));
			fInfo.fileType = 'PNGf';
			
#if defined(__MWERKS__)
# if TARGET_API_MAC_CARBON
			if (!_isOSX())
#endif
				fInfo.fileCreator = 'ogle';
#endif
			memcpy(&catInfo.finderInfo, &fInfo, sizeof(fInfo));
			FSCreateFileUnicode(&theFSRef, uname.length, uname.unicode, kFSCatInfoFinderInfo, &catInfo, NULL, &theFSSpec);
		}
	}

	if ( FSpMakeFSRef(&theFSSpec, &theFSRef) == noErr )
	{
		thePath = (char *)malloc(PNGUMAXPATH);
		if (thePath)
		{
#if defined(__GNUC__)
			if ( FSRefMakePath(&theFSRef, (UInt8 *)thePath, PNGUMAXPATH) == noErr )
				ansiFile = fopen((const char*)thePath, fileMode);
#elif defined(__MWERKS__)
			ansiFile = FSRef_fopen(&theFSRef, fileMode);
#endif
			free(thePath);
		}
	}
		
	return ansiFile;
}


/* ---------------------------------------------------------------------------
	BuildImageDataFromPicture
	
	Use Quartz on Mac OS X, Quickdraw on Mac OS 9
---------------------------------------------------------------------------- */
Boolean BuildImageDataFromPicture(REALpicture thePicture, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData)
{
#if TARGET_RT_MAC_MACHO
	return _BuildImageDataFromPictureCG(thePicture, theWidth, theHeight, imageData);
#else
# if TARGET_API_MAC_CARBON
	if (_isOSX() && machFuncsAvailable)
		return _BuildImageDataFromPictureCG(thePicture, theWidth, theHeight, imageData);
	else
# endif
		return _BuildImageDataFromPictureQD(thePicture, theWidth, theHeight, imageData);
#endif
	return false;
}


/* ---------------------------------------------------------------------------
	BuildImageDataFromPictureAndMask
	
	Use Quartz on Mac OS X, Quickdraw on Mac OS 9
---------------------------------------------------------------------------- */
Boolean BuildImageDataFromPictureAndMask(REALpicture thePicture, REALpicture theMask, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData)
{	
#if TARGET_RT_MAC_MACHO
	return _BuildImageDataFromPictureAndMaskCG(thePicture, theMask, theWidth, theHeight, imageData);
#else
# if TARGET_API_MAC_CARBON
	if (_isOSX() && machFuncsAvailable)
		return _BuildImageDataFromPictureAndMaskCG(thePicture, theMask, theWidth, theHeight, imageData);
	else
# endif
		return _BuildImageDataFromPictureAndMaskQD(thePicture, theMask, theWidth, theHeight, imageData);
#endif
	return false;
}


#pragma mark -


#if TARGET_API_MAC_CARBON

/* ---------------------------------------------------------------------------
	_BuildImageDataFromPictureCG
	
	Quartz path for Mac OS X (requires 10.2 or better)
	
	FIXME? Making some assumptions on the storage format...
	This expects a 32bit Picture and assumes 8bpp ARGB. I'm not sure how
	this is stored on Intel, but since it's a coerced Quickdraw context it 
	may just work as-is...
---------------------------------------------------------------------------- */
static Boolean _BuildImageDataFromPictureCG(REALpicture thePicture, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData)
{	
	UInt32					x, y, srcRowBytes, dstRowBytes;
	size_t					srcbpp;
	UInt8					*srcPtr = NULL, *dstPtr;
	Boolean					success = false;
	CGContextRef			cgContext = NULL;
	REALpictureDescription	desc;

	// Currently the best we can get is a GWorld, but we can deal with that...
	if ( !REALLockPictureDescription( thePicture, &desc, pictureMacintoshGWorld )) return false;
		
	if ( QDBeginCGContext( (CGrafPtr)desc.pictureData, &cgContext ) == noErr )
	{
		srcPtr = (UInt8 *)CGBitmapContextGetData(cgContext);
		if (srcPtr)
		{
			// Paranoid...
			if ( ( CGBitmapContextGetWidth(cgContext) == theWidth ) ||
				 ( CGBitmapContextGetHeight(cgContext) == theHeight ) )
			{	
				dstPtr = imageData;
				dstRowBytes = (theWidth * 3);
				srcRowBytes = CGBitmapContextGetBytesPerRow( cgContext );
				srcbpp = ( CGBitmapContextGetBitsPerPixel( cgContext ) + 7 ) / 8;
				
				for (y = 0; y < theHeight; y++)
				{
					// Copy another scanline
					for (x = 0; x < theWidth; x++)
					{
						dstPtr[(x * 3) + 0] = srcPtr[(x * srcbpp) + 1];
						dstPtr[(x * 3) + 1] = srcPtr[(x * srcbpp) + 2];
						dstPtr[(x * 3) + 2] = srcPtr[(x * srcbpp) + 3];
					}

					// Move on to the next scanline
					srcPtr += srcRowBytes;
					dstPtr += dstRowBytes;
				}
				
				success = true;
			}
		}
		QDEndCGContext( (CGrafPtr)desc.pictureData, &cgContext );
	}
	REALUnlockPictureDescription( thePicture );
	
	return success;
}


/* ---------------------------------------------------------------------------
	_BuildImageDataFromPictureAndMaskCG
	
	Quartz path for Mac OS X (requires 10.2 or better)
	
	FIXME? Making some assumptions on the storage format...
	This expects a 32bit Picture and assumes 8bpp ARGB. I'm not sure how
	this is stored on Intel, but since it's a coerce Quickdraw context it 
	may just work as-is...
---------------------------------------------------------------------------- */
static Boolean _BuildImageDataFromPictureAndMaskCG(REALpicture thePicture, REALpicture theMask, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData)
{
	UInt32					x, y, srcRowBytes, dstRowBytes;
	size_t					srcbpp;
	UInt8					*srcPtr = NULL, *dstPtr;
	Boolean					success = false;
	CGContextRef			cgContext = NULL;
	REALpictureDescription	desc;

	// Currently the best we can get is a GWorld, but we can deal with that...
	if ( !REALLockPictureDescription( thePicture, &desc, pictureMacintoshGWorld )) return false;
			
	if ( QDBeginCGContext( (CGrafPtr)desc.pictureData, &cgContext ) == noErr )
	{
		srcPtr = (UInt8 *)CGBitmapContextGetData(cgContext);
		if (srcPtr)
		{
			// Paranoid...
			if ( ( CGBitmapContextGetWidth(cgContext) == theWidth ) ||
				 ( CGBitmapContextGetHeight(cgContext) == theHeight ) )
			{
				dstPtr = imageData;
				dstRowBytes = (theWidth * 4);
				srcRowBytes = CGBitmapContextGetBytesPerRow( cgContext );
				srcbpp = ( CGBitmapContextGetBitsPerPixel( cgContext ) + 7 ) / 8;
				
				for (y = 0; y < theHeight; y++)
				{
					// Copy another scanline
					for (x = 0; x < theWidth; x++)
					{
						dstPtr[(x * 4) + 0] = srcPtr[(x * srcbpp) + 1];
						dstPtr[(x * 4) + 1] = srcPtr[(x * srcbpp) + 2];
						dstPtr[(x * 4) + 2] = srcPtr[(x * srcbpp) + 3];
						dstPtr[(x * 4) + 3] = 255;
					}

					// Move on to the next scanline
					srcPtr += srcRowBytes;
					dstPtr += dstRowBytes;
				}
				
				success = true;
			}
		}
		QDEndCGContext( (CGrafPtr)desc.pictureData, &cgContext );
	}
	REALUnlockPictureDescription( thePicture );
	
	// And now the mask...
	
	if (success)
	{
		success = false;
		if ( !REALLockPictureDescription( theMask, &desc, pictureMacintoshGWorld )) return false;
			
		if ( QDBeginCGContext( (CGrafPtr)desc.pictureData, &cgContext ) == noErr )
		{
			srcPtr = (UInt8 *)CGBitmapContextGetData(cgContext);
			if (srcPtr)
			{
				// Paranoid...
				if ( ( CGBitmapContextGetWidth(cgContext) == theWidth ) ||
					 ( CGBitmapContextGetHeight(cgContext) == theHeight ) )
				{
					dstPtr = &imageData[0];
					dstRowBytes = (theWidth * 4);
					srcRowBytes = CGBitmapContextGetBytesPerRow( cgContext );
					srcbpp = ( CGBitmapContextGetBitsPerPixel( cgContext ) + 7 ) / 8;
					
					for (y = 0; y < theHeight; y++)
					{
						// Fill in the alpha byte
						// FIXME? Rb2006 (and maybe 2005) seem to properly fill in the alpha byte
						// in the data storage - Rb5.5 does not, so we use an inverted R component.
						// This assumes the mask is greyscale for proper results, but perhaps we
						// should convert the RGB components into luminance values instead...
						for (x = 0; x < theWidth; x++)
							dstPtr[(x * 4) + 3] = 255-srcPtr[(x * srcbpp) + 1];

						// Move on to the next scanline
						srcPtr += srcRowBytes;
						dstPtr += dstRowBytes;
					}
					
					success = true;
				}
			}
			QDEndCGContext( (CGrafPtr)desc.pictureData, &cgContext );
		}
		REALUnlockPictureDescription( theMask );
	}
	
	return success;
}

#endif// TARGET_API_MAC_CARBON

#if !TARGET_RT_MAC_MACHO

/* ---------------------------------------------------------------------------
	_BuildImageDataFromPictureQD
	
	Build a block of RGB data from a Picture (Quickdraw)
	Modified from LoadSaveJPEG (no temp storage).
	
	FIXME? This assumes the Picture storage is 32bit 8bbp ARGB
---------------------------------------------------------------------------- */
static Boolean _BuildImageDataFromPictureQD(REALpicture thePicture, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData)
{	
	UInt32					x, y, srcRowBytes, dstRowBytes, srcbpp;
	UInt8					*srcPtr, *dstPtr;
	PixMapHandle    		thePixMap;
	REALpictureDescription	desc;
	Boolean					success = false;

	if ( !REALLockPictureDescription( thePicture, &desc, pictureMacintoshGWorld )) return false;
    thePixMap = GetGWorldPixMap( (GWorldPtr)desc.pictureData );

    if ( thePixMap )
    {
	    if ( LockPixels(thePixMap) )
		{
			srcPtr = (UInt8 *)GetPixBaseAddr(thePixMap);
			
			if ( srcPtr )
			{
				dstPtr = imageData;
				srcRowBytes = (*thePixMap)->rowBytes & 0x7FFF;
				srcbpp = (UInt32((*thePixMap)->pixelSize) + 7) / 8;
				dstRowBytes = (theWidth * 3);
				
				for (y = 0; y < theHeight; y++)
				{
					// Copy another scanline
					for (x = 0; x < theWidth; x++)
					{
						dstPtr[(x * 3) + 0] = srcPtr[(x * srcbpp) + 1];
						dstPtr[(x * 3) + 1] = srcPtr[(x * srcbpp) + 2];
						dstPtr[(x * 3) + 2] = srcPtr[(x * srcbpp) + 3];
					}

					// Move on to the next scanline
					srcPtr += srcRowBytes;
					dstPtr += dstRowBytes;
				}
				
				success = true;
			}
			
			UnlockPixels(thePixMap);
		}
	}
	
	REALUnlockPictureDescription( thePicture );
	
	return success;
}


/* ---------------------------------------------------------------------------
	_BuildImageDataFromPictureAndMaskQD
	
	Build a block of RGBA data from a Picture (Quickdraw)
	Modified from LoadSaveJPEG (no temp storage, fill alpha byte).
	
	FIXME? This assumes the Picture storage is 32bit 8bbp ARGB
---------------------------------------------------------------------------- */
static Boolean _BuildImageDataFromPictureAndMaskQD(REALpicture thePicture, REALpicture theMask, UInt32 theWidth, UInt32 theHeight, UInt8 *imageData)
{	
	UInt32					x, y, srcRowBytes, dstRowBytes, srcbpp;
	UInt8					*srcPtr, *dstPtr;
	PixMapHandle    		thePixMap;
	REALpictureDescription	desc;
	Boolean					success = false;

	if ( !REALLockPictureDescription( thePicture, &desc, pictureMacintoshGWorld )) return false;
    thePixMap = GetGWorldPixMap( (GWorldPtr)desc.pictureData );

    if ( thePixMap )
    {
	    if ( LockPixels(thePixMap) )
		{
			srcPtr = (UInt8 *)GetPixBaseAddr(thePixMap);
			
			if ( srcPtr )
			{
				dstPtr = imageData;
				srcRowBytes = (*thePixMap)->rowBytes & 0x7FFF;
				srcbpp = (UInt32((*thePixMap)->pixelSize) + 7) / 8;
				dstRowBytes = (theWidth * 4);
				
				for (y = 0; y < theHeight; y++)
				{
					// Copy another scanline
					for (x = 0; x < theWidth; x++)
					{
						dstPtr[(x * 4) + 0] = srcPtr[(x * srcbpp) + 1];
						dstPtr[(x * 4) + 1] = srcPtr[(x * srcbpp) + 2];
						dstPtr[(x * 4) + 2] = srcPtr[(x * srcbpp) + 3];
						dstPtr[(x * 4) + 3] = 255;
					}

					// Move on to the next scanline
					srcPtr += srcRowBytes;
					dstPtr += dstRowBytes;
				}
				
				success = true;
			}
			
			UnlockPixels(thePixMap);
		}
	}
	
	REALUnlockPictureDescription( thePicture );
	
	// And now the mask...
	
	if ( success )
	{
		success = false;
		thePixMap = NULL;
		if ( !REALLockPictureDescription( theMask, &desc, pictureMacintoshGWorld )) return false;
	    thePixMap = GetGWorldPixMap( (GWorldPtr)desc.pictureData );

	    if ( thePixMap )
	    {
		    if ( LockPixels(thePixMap) )
			{
				srcPtr = (UInt8 *)GetPixBaseAddr(thePixMap);
				
				if ( srcPtr )
				{
					dstPtr = &imageData[0];
					srcRowBytes = (*thePixMap)->rowBytes & 0x7FFF;
					srcbpp = (UInt32((*thePixMap)->pixelSize) + 7) / 8;
					dstRowBytes = (theWidth * 4);
					
					for (y = 0; y < theHeight; y++)
					{
						// Fill in the alpha byte
						for (x = 0; x < theWidth; x++)
							dstPtr[(x * 4) + 3] = srcPtr[(x * srcbpp)];

						// Move on to the next scanline
						srcPtr += srcRowBytes;
						dstPtr += dstRowBytes;
					}

					success = true;
				}
				
				UnlockPixels(thePixMap);
			}
		}
		
		REALUnlockPictureDescription( theMask );
	}	
	
	return success;
}

#endif// !TARGET_RT_MAC_MACHO

/* ---------------------------------------------------------------------------
	_isOSX
	
	Determines if we're in OSX for CFM Carbon environment.
---------------------------------------------------------------------------- */
# if defined(__MWERKS__) && TARGET_API_MAC_CARBON
static SInt32 _isOSX(void)
{
	UInt32 response;

	Gestalt(gestaltSystemVersion, (SInt32 *)&response);
	if (response >= 0x01000) return true;
	return false;
}
# endif
